#!BPY

"""
 Name: 'PMD file for MMD(.pmd) ...'
 Blender: 248a
 Group: 'Export'
 Tooltip: 'Export PMD file for MikuMikuDance.'
"""

__author__= "BouKiCHi_&NANAshi"
__version__= "2.2"

__bpydoc__=""

from Blender import *
import bpy
import copy
import re
import os

from xml.dom.minidom import parse
from struct import *

import pmd_2

CHENGE_MATRIX = Mathutils.Matrix([1,0,0,0],[0,1,0,0],[0,0,-1,0],[0,0,0,1])
DEBUG = True

#
# Log
#

def SetLog(message):
	if DEBUG:
		print message	

#
# Take XML
#

#GetXMLText
def getXMLText(basenode):
	nodelist = basenode.childNodes
	for node in nodelist:
		if node.nodeType == node.TEXT_NODE:
			return node.data.encode('UTF-8')
	return ""

def getXMLAttribute(basenode,attrname):	
	attr = basenode.getAttributeNode(attrname)	
	if attr != None :
		return attr.nodeValue.encode('UTF-8')
	return "0"

def TakeXML(filename):

	PMD_Title = "None"
	PMD_Comment = "None"
	PMD_Faces = {}
	PMD_Faces_index = {}
	PMD_Labels = {}
	PMD_Labels_index ={}
	PMD_Bone_Status = {}
	PMD_IK = {}
	PMD_IK_index = {}
	PMD_XML_DATA = {}

	xmlPath = os.path.splitext(filename)[0] + '.xml'
	
	if os.access(xmlPath, os.F_OK):
		obj = parse(filename) 
	
		# GetTitle
		nodelist = obj.getElementsByTagName("title")
		for node in nodelist:
			PMD_Title = getXMLText(node)
			SetLog("Title=" + PMD_Title)
	
		#GetComment
		nodelist = obj.getElementsByTagName("comment")
		for node in nodelist:
			PMD_Comment = getXMLText(node)
			SetLog("Comment=" + PMD_Comment)
		
		#GetFaceList
		element_array = obj.getElementsByTagName("expression")
		for element in element_array:
			nodelist = element.getElementsByTagName("face")
			for node in nodelist:
				face_index = getXMLAttribute(node,"index")
				face_group = getXMLAttribute(node,"group")
				face_name = getXMLText(node)
				PMD_Faces[face_name] = int(face_group)
				PMD_Faces_index[face_index] = face_name
		
		#GetBoneList
		element_array = obj.getElementsByTagName("labels")
		for element in element_array:
			nodelist = element.getElementsByTagName("label")
			for node in nodelist:
				labelname = getXMLAttribute(node,"name")
				labelindex = getXMLAttribute(node,"index")
				SetLog("Label > %s" % labelname)
				PMD_Labels[labelname] = {};
				PMD_Labels_index[labelindex] = labelname
				bones = node.getElementsByTagName("bone")
				for bone in bones:
					blender_bone_name = getXMLAttribute(bone,"blenderName")
					pmd_bone_name = getXMLText(bone)					
					pmd_bone_kind = int(getXMLAttribute(bone,"kind"))					
					label_index = getXMLAttribute(bone,"index")
					PMD_Bone_Status[blender_bone_name] = (pmd_bone_name,pmd_bone_kind)
					PMD_Labels[labelname][label_index] = blender_bone_name
				if labelname == "Hidden":
					del(PMD_Labels["Hidden"])
		#GrtIKList
		element_array = obj.getElementsByTagName("IKbones")
		for element in element_array:
			nodelist = element.getElementsByTagName("IKbone")
			for node in nodelist:
				IK_bone_name = getXMLText(node)
				IK_index = getXMLAttribute(node,"index")
				IK1 = int(getXMLAttribute(node,"IK1"))
				IK2 = float(getXMLAttribute(node,"IK2"))
				IK_parent =  getXMLAttribute(node,"parent")
				PMD_IK[IK_bone_name] = (IK_parent,IK1,IK2)
				PMD_IK_index[IK_index] = IK_bone_name
		
	#setDefaule
	PMD_Header = ['',0.0,'','']
	PMD_Header[0] = 'Pmd'
	PMD_Header[1] = 1.0
	PMD_Header[2] = PMD_Title
	PMD_Header[3] = PMD_Comment
	
	PMD_XML_DATA["PMD_Faces"] = PMD_Faces
	PMD_XML_DATA["PMD_Faces_index"] = PMD_Faces_index
	PMD_XML_DATA["PMD_Labels"] = PMD_Labels
	PMD_XML_DATA["PMD_Labels_index"] = PMD_Labels_index
	PMD_XML_DATA["PMD_Bone_Status"] = PMD_Bone_Status
	PMD_XML_DATA["PMD_IK"] = PMD_IK
	PMD_XML_DATA["PMD_IK_index"] = PMD_IK_index	
	
	return PMD_Header,PMD_XML_DATA 

#
# Take_Vertices
#
# Vertex&Wait&Poly
#

def TakeVertices(bone_dic):
	pmd_vertices = []
	pmd_weights =[]	
	pmd_poly = []		
	pmd_material = []
	pmd_skins = []
	
	scene = Scene.GetCurrent()
	object_list = scene.objects
	
	base_dic = {}
	base_index = 0
	
	pmd_vertices_index = {}
	
	material_dic = {}
	
	skin_dic = {}
		
	has_key = 0
	
	TEMP_Vertices = []
	TEMP_Faces = {}
	TEMP_ADD_Vertices = []
	TEMP_ADD_Vert_Dic = {}
	TEMP_Keys = {}
	#TEMP_Materials = []
	TEMF_Vertices_Margin = 0

	
	materials = Material.Get()	
	for mat in materials:		
		if mat.users == 0:
			continue
		TEMP_Faces[mat.name] = []
	
	# Get Data	
	for ob in object_list:
		object_matrix = ob.matrixWorld #* CHENGE_MATRIX
		if ob.getType() != "Mesh":
			continue
		mesh = ob.getData(0,1)
		print mesh			
		#mesh = Mesh.Get(ob.getData(1))
		mesh.update()
		if mesh.users == 0:
			continue
			
		SetLog("GET "+ob.name)
		SetLog("Mesh GET "+mesh.name)

		for mesh_vertex in mesh.verts:
						
			# SET VERTEX
			pmd_vort = []		
			pmd_vort[0:] = mesh_vertex.co * object_matrix * CHENGE_MATRIX
			pmd_vort[3:] = (mesh_vertex.no * object_matrix.rotationPart() * CHENGE_MATRIX) #.normalize()
			
			# UV
			pmd_vort[6:] =(0,0)
			
			# ADD VERTEX
			pmd_vertices.append(pmd_vort)
			
			# SET WEIGHT
			pmd_weight = [0,0,100,0]
			temp_bones = mesh.getVertexInfluences(mesh_vertex.index)
			wait_bones = []
			for temp_bone in temp_bones:
				if temp_bone[0] == "no_edge_flag":
					pmd_weight[3] = 1 #no edge
				elif bone_dic.has_key(temp_bone[0]):					
					if temp_bone[1] > 0.01 :
						wait_bones.append(temp_bone)
											
			if len(wait_bones) >= 1:					
				pmd_weight[0] = bone_dic[wait_bones[0][0]] #Bone1.index			
			if len(wait_bones) >= 2:
				pmd_weight[1] = bone_dic[wait_bones[1][0]] #Bone2.index
				pmd_weight[2] = 100 * wait_bones[0][1] / (wait_bones[0][1] + wait_bones[1][1])  #bonewait
			
			#ADD Weight
			pmd_weights.append(pmd_weight)
		
		
		#SET FACE
		# check Face Flip
		is_inverce = -1
		if ob.SizeX < 0.0:
			is_inverce *= -1
		if ob.SizeY < 0.0:
			is_inverce *= -1
		if ob.SizeZ < 0.0:
			is_inverce *= -1
		
		for face in mesh.faces:
			
			if is_inverce > 0 :						
				TEMP_FACE_Verts = []
				TEMP_FACE_Verts.append(face.v[0].index+TEMF_Vertices_Margin)
				TEMP_FACE_Verts.append(face.v[1].index+TEMF_Vertices_Margin)
				TEMP_FACE_Verts.append(face.v[2].index+TEMF_Vertices_Margin)
				
				TEMP_FACE_UVs = []	
				if(mesh.faceUV):
					TEMP_FACE_UVs.append((face.uv[0][0],1-face.uv[0][1]))
					TEMP_FACE_UVs.append((face.uv[1][0],1-face.uv[1][1]))
					TEMP_FACE_UVs.append((face.uv[2][0],1-face.uv[2][1]))
				else:
					TEMP_FACE_UVs.append((0,0))				
					TEMP_FACE_UVs.append((0,0))				
					TEMP_FACE_UVs.append((0,0))				
				
				TEMP_Faces[mesh.materials[face.mat].name].append((TEMP_FACE_Verts,TEMP_FACE_UVs))
				
				# sqwere
				if len(face.v) == 4:
					TEMP_FACE_Verts = []
					TEMP_FACE_Verts.append(face.v[0].index+TEMF_Vertices_Margin)
					TEMP_FACE_Verts.append(face.v[2].index+TEMF_Vertices_Margin)
					TEMP_FACE_Verts.append(face.v[3].index+TEMF_Vertices_Margin)
					
					TEMP_FACE_UVs = []	
					if(mesh.faceUV):
						TEMP_FACE_UVs.append((face.uv[0][0],1-face.uv[0][1]))
						TEMP_FACE_UVs.append((face.uv[2][0],1-face.uv[2][1]))
						TEMP_FACE_UVs.append((face.uv[3][0],1-face.uv[3][1]))
					else:
						TEMP_FACE_UVs.append((0,0))				
						TEMP_FACE_UVs.append((0,0))				
						TEMP_FACE_UVs.append((0,0))	
					
					TEMP_Faces[mesh.materials[face.mat].name].append((TEMP_FACE_Verts,TEMP_FACE_UVs))
				
			else :
				
				TEMP_FACE_Verts = []
				TEMP_FACE_Verts.append(face.v[0].index+TEMF_Vertices_Margin)
				TEMP_FACE_Verts.append(face.v[2].index+TEMF_Vertices_Margin)
				TEMP_FACE_Verts.append(face.v[1].index+TEMF_Vertices_Margin)
				
				TEMP_FACE_UVs = []	
				if(mesh.faceUV):
					TEMP_FACE_UVs.append((face.uv[0][0],1-face.uv[0][1]))
					TEMP_FACE_UVs.append((face.uv[2][0],1-face.uv[2][1]))
					TEMP_FACE_UVs.append((face.uv[1][0],1-face.uv[1][1]))
				else:
					TEMP_FACE_UVs.append((0,0))				
					TEMP_FACE_UVs.append((0,0))				
					TEMP_FACE_UVs.append((0,0))				
				
				TEMP_Faces[mesh.materials[face.mat].name].append((TEMP_FACE_Verts,TEMP_FACE_UVs))
				
				# sqwere
				if len(face.v) == 4:
					TEMP_FACE_Verts = []
					TEMP_FACE_Verts.append(face.v[0].index+TEMF_Vertices_Margin)
					TEMP_FACE_Verts.append(face.v[3].index+TEMF_Vertices_Margin)
					TEMP_FACE_Verts.append(face.v[2].index+TEMF_Vertices_Margin)
					
					TEMP_FACE_UVs = []	
					if(mesh.faceUV):
						TEMP_FACE_UVs.append((face.uv[0][0],1-face.uv[0][1]))
						TEMP_FACE_UVs.append((face.uv[3][0],1-face.uv[3][1]))
						TEMP_FACE_UVs.append((face.uv[2][0],1-face.uv[2][1]))
					else:
						TEMP_FACE_UVs.append((0,0))				
						TEMP_FACE_UVs.append((0,0))				
						TEMP_FACE_UVs.append((0,0))	
					
					TEMP_Faces[mesh.materials[face.mat].name].append((TEMP_FACE_Verts,TEMP_FACE_UVs))		
				
		#SET KEY
		if mesh.key != None :
			if mesh.key.type == Key.Types.MESH:
				SetLog("Key GET")
				# Save Key
				for key_block in mesh.key.blocks:
					if not TEMP_Keys.has_key(key_block.name):
						TEMP_Keys[key_block.name] = {}
					temp_block = key_block.getData()
					for key_index in range(len(temp_block)):
						TEMP_Keys[key_block.name][key_index+TEMF_Vertices_Margin] = temp_block[key_index] * object_matrix * CHENGE_MATRIX
						#* object_matrix.rotationPart() * CHENGE_MATRIX  
	
		TEMF_Vertices_Margin = len(pmd_vertices)
	
	#SET Data
	for TEMP_Face_Mat_Key,TEMP_Face_data in TEMP_Faces.items():
		for TEMP_Face_index in range(len(TEMP_Face_data)):
			TEMP_Face = TEMP_Face_data[TEMP_Face_index]
			for TEMP_Face_Vert_index,TEMP_Face_Vert_UV,TEMP_Face_Vert_index_index in zip(TEMP_Face[0],TEMP_Face[1],range(len(TEMP_Face[0]))):
				if pmd_vertices[TEMP_Face_Vert_index][6:8] == [0,0]:
					pmd_vertices[TEMP_Face_Vert_index][6:8] = TEMP_Face_Vert_UV
				elif pmd_vertices[TEMP_Face_Vert_index][6:8] != [TEMP_Face_Vert_UV[0],TEMP_Face_Vert_UV[1]]:
					vetex_check_key = (pmd_vertices[TEMP_Face_Vert_index][0],pmd_vertices[TEMP_Face_Vert_index][1],pmd_vertices[TEMP_Face_Vert_index][2],TEMP_Face_Vert_UV[0],TEMP_Face_Vert_UV[1])
					if TEMP_ADD_Vert_Dic.has_key(vetex_check_key):
						TEMP_Faces[TEMP_Face_Mat_Key][TEMP_Face_index][0][TEMP_Face_Vert_index_index] = TEMP_ADD_Vert_Dic[vetex_check_key]
					else :
						newNo = len(pmd_vertices)
						newVert = pmd_vertices[TEMP_Face_Vert_index][:]
						newVert[6:8] = TEMP_Face_Vert_UV
						#ADD Vert
						pmd_vertices.append(newVert)
						pmd_weights.append(pmd_weights[TEMP_Face_Vert_index][:])
						#ADD Dic
						TEMP_ADD_Vert_Dic[vetex_check_key] = newNo
						TEMP_Faces[TEMP_Face_Mat_Key][TEMP_Face_index][0][TEMP_Face_Vert_index_index] = newNo
						#ADD Key
						for TEMP_KEY_data in TEMP_Keys.values():
							if TEMP_KEY_data.has_key(TEMP_Face_Vert_index):
								TEMP_KEY_data[newNo] = TEMP_KEY_data[TEMP_Face_Vert_index][:]
						
			
#			for face in mesh.faces:
#				TEMP_Faces.append(face)
#			
#			print len(TEMP_Faces)
#			
#			if mesh.hasFaceUV():	#hasUV				
#				SetLog("HAS UVTex :" + mesh.name)
#				for face in mesh.faces:
#					#UV get		
#					TMP_face_vert = []
#					for vert_index in range(len(face.v)):
#						vert = face.v[vert_index]
#						uv = face.uv[vert_index]
#						vetex_check_key = (vert.co[0],vert.co[1],vert.co[2],uv[0],uv[1])
#						new_vert_index = vert.index 
#						 											
#						if vertex_link.has_key(vert.index):
#							if vertex_check.has_key(vetex_check_key):
#								new_vert_index = vertex_check[vetex_check_key]								
#								TMP_face_vert.append(mesh.verts[new_vert_index])
#							else:								
#								# create vertex								
#								vertex_add += 1
#								old_vert_index = vert.index	
#								new_vert_index = len(mesh.verts) 
#								mesh.verts.append(NMesh.Vert(vert.co[0],vert.co[1],vert.co[2]))
#								new_vert = mesh.verts[new_vert_index]
#								new_vert.no[0:3] = vert.no[0:3]
#								
#								vertex_check[vetex_check_key] = new_vert_index
#								vertex_link[new_vert_index] = old_vert_index							
#								TMP_face_vert.append(new_vert)
#						else:
#							vertex_link[new_vert_index] = new_vert_index
#							vertex_check[vetex_check_key] = new_vert_index							
#							TMP_face_vert.append(vert)
#												
#					#SET NEW FACE		
#					face.v = TMP_face_vert
#				mesh.update()								
#				SetLog("Add Vertex = %d" % vertex_add)
#								
#				# wait Assign
#				for new_vert_index in vertex_link:
#					old_vert_index = vertex_link[new_vert_index]
#					if new_vert_index == old_vert_index:
#						continue 
#					wait_bones = mesh.getVertexInfluences(old_vert_index)
#					for wait_bone in wait_bones:
#						mesh.assignVertsToGroup(wait_bone[0],[new_vert_index],wait_bone[1],"add") # Fail ?
#				
#				# key repere
#				if mesh.key != None and vertex_add > 0:
#					SetLog("Key RESET")				
#					TempKeys = {}
#					TempKeysTag = []
#					if mesh.key.type == Key.Types.MESH:
#						# Save Key
#						for key_block in mesh.key.blocks:
#							TempKeys[key_block.name] = key_block.getData()
#							TempKeysTag.append(key_block.name)						 
#						# Delete Key
#						mesh.removeAllKeys()
#						# Make Key
#						mesh.insertKey()					
#						for TempKeyTag in TempKeysTag:
#							if(TempKeyTag == "Basis"):
#								# Basis
#								continue												
#							for mesh_vert in mesh.verts:							
#								if mesh_vert.index not in vertex_link.keys():
#									key_vert = TempKeys[TempKeyTag][mesh_vert.index]
#								else:
#									key_vert = TempKeys[TempKeyTag][vertex_link[mesh_vert.index]]
#								
#								mesh_vert[0:3] = key_vert[0:3]
#																
#							# Add Key
#							mesh.update()
#							mesh.insertKey()
#							# Key Rename 			
#							obj = mesh.key.blocks.pop()
#							obj.name = TempKeyTag
#							mesh.key.blocks.append(obj)
#									
#						# Restore
#						#for mesh_vert,key_vert in zip(mesh.verts,TempKeys["Basis"]):
#						for mesh_vert in mesh.verts:							
#							if mesh_vert.index not in vertex_link.keys():
#								key_vert = TempKeys["Basis"][mesh_vert.index]
#							else:
#								key_vert = TempKeys["Basis"][vertex_link[mesh_vert.index]]
#							mesh_vert[0:3] = key_vert[0:3]					
#						mesh.update()
#						
#	# Get Mesh
#	for ob in object_list:
#		
#		object_matrix = ob.matrixWorld #* CHENGE_MATRIX		
#		
#		if ob.getType() == "Mesh":
#			SetLog("MeshObject = %s" % ob.name)
#			mesh = ob.getData(0,0)
#			if mesh == None:
#				SetLog("no Mesh Data")
#				continue  # not Mesh
#			mesh.update()
#			
#			# SetIndex
#			pmd_vertices_index[mesh.name] = len(pmd_vertices)			
#			
#			# Get Vertex				
#			for mesh_vertex in mesh.verts:
#							
#				# SET VERTEX
#				pmd_vort = []		
#				pmd_vort[0:] = mesh_vertex.co * object_matrix * CHENGE_MATRIX
#				pmd_vort[3:] = (mesh_vertex.no * object_matrix.rotationPart() * CHENGE_MATRIX).normalize()
#				pmd_vort[6:] = mesh_vertex.uvco 
#				pmd_vort[7] = pmd_vort[7] * -1 #inverse
#				
#				# Reset UV
#				pmd_vort[6] =0
#				pmd_vort[7] =0		
#							
#				pmd_vertices.append(pmd_vort)
#				
#				# SET WEIGHT
#				pmd_weight = [0,0,100,0]
#				temp_bones = mesh.getVertexInfluences(mesh_vertex.index)
#				wait_bones = []
#				for temp_bone in temp_bones:
#					if temp_bone[0] == "no_edge_flag":
#						pmd_weight[3] = 1
#					elif temp_bone[1] > 0.01 :
#						wait_bones.append(temp_bone)
#												
#				if len(wait_bones) >= 1:					
#					pmd_weight[0] = bone_dic[wait_bones[0][0]] #Bone1.index			
#				if len(wait_bones) >= 2:
#					pmd_weight[1] = bone_dic[wait_bones[1][0]] #Bone2.index
#					pmd_weight[2] = 100 * wait_bones[0][1] / (wait_bones[0][1] + wait_bones[1][1])  #bonewait
#									  
#				#pmd_weight[3] = 0 #unkown Flag
#			
#				pmd_weights.append(pmd_weight)
#				
	#SET VERTEX KEY
	if TEMP_Keys.has_key("Basis"):	
		for TEMP_KEY_Name in TEMP_Keys.keys():
			skin_dic[TEMP_KEY_Name] = []
			
		for TEMP_KEY_data in TEMP_Keys.items():
			if TEMP_KEY_data[0] == "Basis":
				continue
			for key,data in TEMP_KEY_data[1].items():							
				if data != TEMP_Keys["Basis"][key]:
					if not base_dic.has_key(key):
						base_dic[key] = len(base_dic)
					gap = [0,0,0,0]
					gap[0] = base_dic[key]
					gap[1:] = (data - TEMP_Keys["Basis"][key])
					skin_dic[TEMP_KEY_data[0]].append(gap)

		## base_dic
		pmd_skin = ["Basis",0,()]
		#pmd_skin[0] = key_block.name
		base_dic_arr = [(0,0,0,0)] * len(base_dic)		
		for base_dic_key in base_dic.keys():
			base_dic_vart = [0,0,0,0]
			base_dic_vart[0] = base_dic_key
			base_dic_vart[1:] = pmd_vertices[base_dic_key][0:3]
			base_dic_arr[base_dic[base_dic_key]] = base_dic_vart
		pmd_skin[2] = base_dic_arr	
		pmd_skins.append(pmd_skin)
		## faces		
		for skin_key in skin_dic.keys():
			if  skin_key == "Basis":
				continue			
			pmd_skin = [skin_key,1,skin_dic[skin_key]]
			pmd_skins.append(pmd_skin)
			
		
	# Get Material & Face	
	materials = Material.Get()	
	for mat in materials:
		if mat.users == 0:
			continue		
		SetLog("Material = %s" % mat.name)
		
		#Material Add
		data = [0.0, 0.0, 0.0, 0.0, 1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0, 0, 0, '']
		(data[0],data[1],data[2]) = mat.getRGBCol()  #deffuse
		(data[3]) = mat.getAlpha()                   #Alpha 					
		(data[4]) = mat.getHardness()                #Shines
		(data[5],data[6],data[7]) = mat.getSpecCol() #Speculer
		(data[8],data[9],data[10]) = mat.getMirCol() #Mirror color()
		data[11] = 0
		data[12] = 1
				
		poly_start_index = len(pmd_poly)
		for face_data in TEMP_Faces[mat.name]:
			pmd_poly.extend(face_data[0])
									
		data[13] = len(pmd_poly) - poly_start_index		
		l_tex = mat.getTextures()
		for tex in l_tex:
			if tex == None:
				continue
			if tex.tex.image != None:
				data[14] = sys.basename(tex.tex.image.getFilename())

		pmd_material.append(data)

	return 	pmd_vertices,pmd_weights,pmd_poly,pmd_material,pmd_skins
	
#
# Take_Bone
#
def SetBone(bone_name,arm_status):
			
	bone = arm_status["arm_data"].bones[bone_name]
	bone_status = arm_status["pmd_bone_status"].get(bone_name,(bone_name,0))
	
	pmd_bone = [0,0,0,0,0,0,0,0]
	pmd_bone[0] = bone_status[0]	#name
	pmd_bone[1] = 65535   			#parent
	pmd_bone[2] = 0     			#to
	pmd_bone[3] = bone_status[1]	#kind
	pmd_bone[4] = 0     			#knum
	pmd_bone[5:] = (bone.head['ARMATURESPACE']*arm_status["arm_mat"])[0:3] #pos
	
	# Check Parent
	if bone.hasParent():
		# set parent
		if bone.parent.name in arm_status["pmd_id"].keys():
			pmd_bone[1] = arm_status["pmd_id"][bone.parent.name]
		else:
			pmd_bone[1] = SetBone(bone.parent.name,arm_status)			

		if(Armature.CONNECTED in bone.options ):
			#set parent "to"
			arm_status["pmd_bones"][pmd_bone[1]][2] = len(arm_status["pmd_bones"])
			
#	# Check IK Parent
#	if bone_name in arm_status["pmd_ik"].keys():
#		pmd_ik = arm_status["pmd_ik"][bone_name]
#		if pmd_ik[0] != "World" and pmd_ik[0] not in arm_status["pmd_id"].keys():
#			SetBone(pmd_ik[0],arm_status)
			
	SetLog(bone_name.decode('UTF-8').encode('shift_jis'))

	next_id = len(arm_status["pmd_bones"])
	arm_status["pmd_id"][bone.name] = next_id
	arm_status["pmd_bones"][next_id] = pmd_bone

	return next_id # arm_status["pmd_id"][bone.name]
	
def TakeBone(PMD_XML_DATA):

	#arm_status["pmd_iks"] = []

	# Get Arms
	arm_objects = []
	objects = Object.Get()
	for ob in objects:
		if ob.type == 'Armature':
			arm_objects.append(ob)
			
	# Set Bones
	PMD_Bone_Status = PMD_XML_DATA["PMD_Bone_Status"]		
	arm_status = {}	
	arm_status["pmd_id"] = {}
	arm_status["pmd_bones"] = {}
	arm_status["pmd_ik"] = PMD_XML_DATA["PMD_IK"]
	
	
	for arm_obj in arm_objects:
		arm_mat = arm_obj.matrixWorld * CHENGE_MATRIX
		arm_data = arm_obj.getData()
		arm_pose = arm_obj.getPose()
		bone_names = arm_pose.bones.keys()
		
		arm_status["arm_mat"] = arm_mat
		arm_status["arm_data"] = arm_obj.getData()
		arm_status["pmd_bone_status"] = PMD_Bone_Status
		#arm_status["arm_pose"] = arm_obj.getPose()
		
		bone_names.sort()
		for bone_name in bone_names:
			if not arm_status["pmd_id"].has_key(bone_name):
				SetBone(bone_name,arm_status)
			
	# Set tail
	#tail_dec ={}
	for bone_name in arm_status["pmd_id"]:
		bone_id = arm_status["pmd_id"][bone_name]
		bone_pmd = arm_status["pmd_bones"][bone_id]
		if bone_pmd[2] == 0 and bone_pmd[3] not in (6,7) :			
			bone = arm_status["arm_data"].bones[bone_name]
			bone_status = arm_status["pmd_bone_status"].get(bone_name,(bone_name,0))
			
			pmd_bone = [0,0,0,0,0,0,0,0]
			pmd_bone[0] = bone_status[0] + "_tail"	#name
			pmd_bone[1] = bone_id			#parent
			pmd_bone[2] = 0     			#to
			pmd_bone[3] = 7					#kind
			pmd_bone[4] = 0     			#knum
			pmd_bone[5:] = (bone.tail['ARMATURESPACE']*arm_status["arm_mat"])[0:3] #pos
		
			next_id = len(arm_status["pmd_bones"]) 
			arm_status["pmd_bones"][next_id] = pmd_bone
			# set parent "to"
			arm_status["pmd_bones"][bone_id][2] = next_id 
			
	# Set IK	
	PMD_IK = PMD_XML_DATA["PMD_IK"]
	PMD_IK_index = PMD_XML_DATA["PMD_IK_index"]	
	pmd_iks = []
			
	ik_List = {}
	for bone_name in arm_status["pmd_id"]:
		bone_constraints = arm_pose.bones[bone_name].constraints
		for bone_constraint in bone_constraints:
			if bone_constraint.type == Constraint.Type.IKSOLVER:
				SetLog("IKset %s" % bone_name)
				IK_bone_id = arm_status["pmd_id"][bone_name]	
				IK_bone =  arm_status["pmd_bones"][IK_bone_id]
				IK_length = bone_constraint[Constraint.Settings.CHAINLEN]
				IK_terget = bone_constraint[Constraint.Settings.TARGET]
		 		IK_terget_name = bone_constraint[Constraint.Settings.BONE]
				if IK_terget_name != None:					
					IK_status = PMD_IK.get(IK_terget_name,("World",20,0.5))
					IK_terget_id = arm_status["pmd_id"][IK_terget_name]
					# SET Status
					arm_status["pmd_bones"][IK_terget_id][3] = 2
					
					IK_list = []
					select_bone_id = IK_bone_id	
					for i in range(IK_length):					
						if select_bone_id == 65535:
							break					
						IK_list.append(select_bone_id)
						select_bone = arm_status["pmd_bones"][select_bone_id]
						arm_status["pmd_bones"][select_bone_id][3] = 4				# Fail case
						arm_status["pmd_bones"][select_bone_id][4] = IK_terget_id
						select_bone_id = select_bone[1]
																				
					# Make IK
					pmd_ik = [0,0,0,0.0,()]
					pmd_ik[0] = IK_terget_id	# IK
					pmd_ik[1] = IK_bone[2]		# IKterget
					pmd_ik[2] = IK_status[1]	# IK_Status1
					pmd_ik[3] = IK_status[2] 	# IK_Status2
					pmd_ik[4] = IK_list
					ik_List[IK_terget_name] = pmd_ik
					#arm_status["pmd_iks"].append(pmd_ik)
		
	ik_keys = PMD_IK_index.keys()
	ik_keys.sort()	
	for ik_key in ik_keys:
		bone_name = PMD_IK_index[ik_key]
		SetLog(bone_name.decode('UTF-8').encode('shift_jis'))
		if ik_List.has_key(bone_name):
		 	pmd_iks.append(ik_List[bone_name])
	for bone_name in ik_List:
		if not PMD_IK.has_key(bone_name):
			pmd_iks.append(ik_List[bone_name])		
						
	return (arm_status["pmd_bones"],pmd_iks,arm_status["pmd_id"])
	
#
# export_pmd
#

def export_pmd(filename):
		
	title = sys.makename(filename,strip = 1)
	SetLog("Filename = %s" % title)
	
	# check Edit Mode
	mode_edit = Window.EditMode() 
	if mode_edit: Window.EditMode(0)

	d_pmd = {}
#	PMD_Faces = {}
#	PMD_Labels = {}
#	PMD_Bone_Status = {}

	savePath = os.path.splitext(filename)[0] + '.pmd'

	#
	# read XML
	#	
	SetLog("XML")	
	xmlPath = os.path.splitext(filename)[0] + '.xml'	
	
	SetLog("Header...")
	d_pmd["Header"],PMD_XML_DATA = TakeXML(xmlPath)

	#
	# Make Armature
	#	d_pmd["Bone"],d_pmd["IK"],bone_dic
	SetLog("Armature...")
	d_pmd["Bone"],d_pmd["IK"],bone_dic = TakeBone(PMD_XML_DATA)
	
	#
	# set Faces to a mesh
	#
	SetLog("Mash & Face...")
	d_pmd["Vertex"],d_pmd["Weight"],d_pmd["Poly"],d_pmd["Material"],d_pmd["Skin"] = TakeVertices(bone_dic)

	
	d_pmd["SkinIndex"] = [] #* len(PMD_Faces)

	#PMD_Faces = d_pmd["Skin"]
	PMD_Faces = PMD_XML_DATA["PMD_Faces"]
	PMD_Faces_index = PMD_XML_DATA["PMD_Faces_index"]
	
	data_index = {}	
	for skin_index in range(len(d_pmd["Skin"])):
		skin = d_pmd["Skin"][skin_index]
		data_index[skin[0]] = skin_index
	
	index_keys = PMD_Faces_index.keys()
	index_keys.sort()
	for index_key in index_keys:
		index_name = PMD_Faces_index[index_key]
		index_name.decode('UTF-8').encode('shift_jis')
		if data_index.has_key(index_name):			
			d_pmd["SkinIndex"].append(data_index[index_name])
			d_pmd["Skin"][data_index[index_name]][1] = PMD_Faces[index_name] # set Face Type
	for skin_index in range(len(d_pmd["Skin"])):		
		index_name = d_pmd["Skin"][skin_index][0]
		if index_name == "Basis":
			continue		
		if not PMD_Faces.has_key(index_name):
			d_pmd["SkinIndex"].append(skin_index)
			d_pmd["Skin"][skin_index][1] = 4 # set Face Type

	
	PMD_Labels = PMD_XML_DATA["PMD_Labels"] 
	PMD_Labels_index = PMD_XML_DATA["PMD_Labels_index"] 
			
	pmd_boneGloup = []
	pmd_boneGloupList = []

	label_index = 1	
	index_keys = PMD_Labels_index.keys()
	index_keys.sort()
	for index_key in index_keys:
		label_name = PMD_Labels_index[index_key]
		if label_name == "Hidden":
			continue
		pmd_boneGloup.append(label_name)				
		bone_keys =  PMD_Labels[label_name].keys()
		bone_keys.sort()
		for bone_key in bone_keys:
			blender_bone_name = PMD_Labels[label_name][bone_key]
			if not bone_dic.has_key(blender_bone_name):
				continue			
			pmd_boneGloupList.append((bone_dic[blender_bone_name],label_index))
		label_index += 1 
		
	d_pmd["DispName"] = pmd_boneGloup
	d_pmd["BoneIndex"] = pmd_boneGloupList
	
	pmd_2.WriteFile(filename,d_pmd)
	
	if mode_edit: Window.EditMode(1)

	print "done."
#
#	if mode_edit: Window.EditMode(1)


#-----------------------------------------------------
if __name__=='__main__':
	
	Window.FileSelector(export_pmd, 'EXPORT PMD', sys.makename(ext='.pmd'))